# coding: utf-8
from builtins import str
from builtins import object

from .database import sqlite

from .utilities import unitConversionFactor, Timer, interpolationWeights

from datetime import datetime
import numpy
import os
import warnings

def _getElementContaining(conn, point):
    """fetch triangle/quad in which requested point lies and
    return node ids of element

    :param conn: connection on a sqlite database
    :type conn: sqlite3.Connection
    :param point: point xy
    :type point: tuple

    :return: mesh element
    :rtype: list
    """
    cur = conn.cursor()
    # fetch triangle/quad in which requested point lies
    project_SRID = str(cur.execute("SELECT srid FROM geometry_columns WHERE f_table_name = 'noeuds'").fetchone()[0])
    cur.execute("SELECT a, b, c, d FROM mailles\
        WHERE Intersects(MakePoint("+",".join([str(point[0]), str(point[1]), project_SRID])+"), \
                         GEOMETRY)\
        LIMIT 1") # the limit 1 is for point at the boundary
    res = cur.fetchone()
    if not res:
        return None
        warnings.warn("Le point "+str(point)+" n'est pas dans le domaine")
        #raise RuntimeError("Le point "+str(point)+" n'est pas dans le domaine")
    return res if res[-1] else res[:-1]# d is null for triangles


class SimpleProvider(object):
    """ Abstract provider """
    PROVIDER_KEY = "simple_provider"

    sourceUnits = {"concentration" : u"kg/m³",
                   "vitesse_darcy_norme" : u"m/s",
                   "potentiel" : u"m"
                   }
    displayName = {"concentration" : "Concentration",
                   "vitesse_darcy_norme" : "Vitesse de Darcy",
                   "potentiel" : "Potentiel"
                   }

    def __init__(self, database, resultColumn, units):
        """Constructor

        :param database: database path
        :type database: string
        :param resultColumn: result type
        :type resultColumn: string
        :param units: units
        :type units: string
        """
        self.__didx = 0
        self.__dates = []

        self.__database = database
        if not self.__database:
            raise RuntimeError("Missing dbname in "+database)
        self.__conn = sqlite.connect(self.__database, "SimpleProvider:init")
        cur = self.__conn.cursor()
        cur.execute("SELECT date FROM dates_simulation ORDER BY id")
        self.setDates([date for [date] in cur.fetchall()])

        self.__nodeValues = []
        self.__nodeSValues = []
        self.__elementValues = []
        self.__elementSValues = []
        self.__unitConvertionFactor = 1.

        self.__resultColumn = resultColumn
        self.__units = units

        self.setResultColumn(resultColumn)

        self.setUnits(units)

        self.setDate(len(self.dates())-1)

    def setDates(self, dates):
        """set list of dates in case node values vary with time

        :param dates: list of dates
        :type dates: list
        """
        self.__dates = dates

    def dates(self):
        """return a list of dates in case node values vary with time

        :return: list of dates
        :rtype: list"""
        return self.__dates

    def setDate(self, didx):
        """Set current date index 

        :param didx: date index
        :type didx: integer
        """
        self.__didx = didx

    def date(self):
        """ Return a current date

        :return: dates list
        :rtype: list
        """
        return self.__didx

    def __updateUnits(self):
        self.setUnits(self.units())

    def contours(self):
        """return a map with the contour (name, group) as key and an array of vertices
        as values, contours are 3D

        :return: contours
        :rtype: dict
        """
        res = {}
        cur = self.__conn.cursor()
        cur.execute("SELECT nom, groupe, AsText(GEOMETRY) FROM contours")
        for nom, group, geom in cur.fetchall():
            res[(nom, group)] = numpy.array(
                    [ [float(c) for c in coord.split()]
                        for coord in geom.
                                replace('LINESTRING Z', '').
                                replace('(', '').
                                replace(')','').
                                split(',') ])
        return res

    def scatters(self):
        """return a map with the contour name as key and an point as value

        :return: scatters
        :rtype: list
        """
        res = []
        cur = self.__conn.cursor()
        cur.execute("""SELECT nom, groupe, X(GEOMETRY), Y(GEOMETRY), altitude FROM points_interet
                union 
                SELECT nom, 'forage', X(POINT), Y(POINT), zsol FROM forages
                """)
        for nom, groupe, x, y, z in cur.fetchall():
            if _getElementContaining(self.__conn, (x,y)):
                res.append({'name':nom, 'point':[x, y, z if z is not None else self.altitudeAt((x,y))], 'group':groupe})
        return res

    def name(self):
        """ Return provider key

        :return: key
        :rtype: string
        """
        return SimpleProvider.PROVIDER_KEY

    def setUnits(self, units):
        """ Set units

        :param units: units
        :type units: string
        """
        sourceUnits = SimpleProvider.sourceUnits[self.resultColumn()]
        if units != sourceUnits:
            if re.match(r'.*Bq/.*', units):
                # get the specific activity in Bq/kg
                cur = self.__conn.cursor()
                cur.execute("""SELECT activite_specifique
                    FROM elements_chimiques WHERE id = (SELECT eid FROM simulations)""")
                activite_specifique = cur.fetchone()[0]
                assert not cur.fetchone() # only one product for a simulation
                self.__unitConvertionFactor = \
                        unitConversionFactor(sourceUnits, units, activite_specifique)
            else:
                self.__unitConvertionFactor = unitConversionFactor(sourceUnits, units)
        else:
            self.__unitConvertionFactor = 1.

    def availableUnits(self):
        """ Return available units

        :return: available units
        :rtype: dict
        """
        return {
            'concentration': [
                u"kg/L", u"g/L", u"mg/L", u"µg/L",
                u"kg/m³", u"g/m³", u"mg/m³", u"µg/m³",
                u"GBq/L", u"MBq/L", u"kBq/L", u"Bq/L", u"mBq/L", u"µBq/L",
                u"GBq/m³", u"MBq/m³", u"kBq/m³", u"Bq/m³", u"mBq/m³", u"µBq/m³"
                ],
            'vitesse_darcy_norme': [
                u"m/s", u"m/s", u"m/an", u"cm/j"
                ],
            'potentiel': [
                u"m"
                ]
            }[self.resultColumn()] if self.resultColumn() else []

    def conversionFactor(self, units=None):
        """conversion from the set units to any units,
        if units is not specify, return in SI units

        :param units: units
        :type units: string
        """

        activite_specifique = 0
        if units and re.match(r'.*Bq/.*', units):
            # get the specific activity in Bq/kg
            cur = self.__conn.cursor()
            cur.execute("""SELECT activite_specifique
                FROM elements_chimiques AS pdt WHERE pdt.id = SELECT eid FROM simulations""")
            activite_specifique = cur.fetchone()[0]
            assert not cur.fetchone() # only one product for a simulation

        sourceUnits = SimpleProvider.sourceUnits[self.resultColumn()]

        return 1./self.__unitConvertionFactor if not units\
                else unitConversionFactor(sourceUnits, units, activite_specifique)\
                     / self.__unitConvertionFactor

    def units(self):
        """ Return current units

        :return: units
        :rtype: string
        """
        return self.__units

    def availableColumns(self):
        """returns a list (display name, colummn name)
        for available results

        :return: available column
        :rtype: list
        """
        available = []
        cur = self.__conn.cursor()
        basename = cur.execute("PRAGMA database_list").fetchone()[2][:-7]
        for columnName in list(SimpleProvider.sourceUnits.keys()):
            binary_file = basename+'.'+columnName.replace('vitesse_darcy_norme', 'darcy')+'.npy'
            if os.path.exists(binary_file):
                available.append((SimpleProvider.displayName[columnName],
                    columnName))
        return available

    def resultColumn(self):
        """ Return current column

        :return: current column
        :rtype: string
        """
        return self.__resultColumn

    def valueAtElement(self):
        """ Return if the mesh is element based

        :return: state
        :rtype: bool
        """
        return str(self.resultColumn()) == 'vitesse_darcy_norme'

    def setResultColumn(self, columnName):
        """ Set the current column
        
        :param columnName: columnName
        :type columnName: string
        """
        # load results since it's the bottleneck for rendering
        reloadNeeded = columnName and (self.resultColumn() != columnName or not len(self.__nodeValues))
        if reloadNeeded:
            self.__elementValues = []
            self.__elementSValues = []
            self.__nodeValues = []
            self.__nodeSValues = []
            cur = self.__conn.cursor()
            basename = cur.execute("PRAGMA database_list").fetchone()[2][:-7]
            binary_file = basename+'.'+columnName.replace('vitesse_darcy_norme', 'darcy')+'.npy'
            pbinary_file = basename+'.s'+columnName.replace('vitesse_darcy_norme', 'darcy')+'.npy'
            if self.valueAtElement():
                if os.path.exists(binary_file):
                    self.__elementValues = numpy.load(binary_file)
                if os.path.exists(pbinary_file):
                    self.__elementSValues = numpy.load(pbinary_file)
            else:
                if os.path.exists(binary_file):
                    self.__nodeValues = numpy.load(binary_file)
                if os.path.exists(pbinary_file):
                    self.__nodeSValues = numpy.load(pbinary_file)
            self.setUnits(self.__units)

    def description(self):
        """ Return description

        :return: description
        :rtype: string
        """
        return "mesh data provider for thyrsis sqlite backend"

    def database(self):
        """ Return db path

        :return: db path
        :rtype: string
        """
        return self.__database

    def nodeCoord(self, zColumn='0'):
        """return a list of coordinates

        :param zColumn: formula for column height or altitude
        :type zColumn: string

        :return: return a list of coordinates
        :rtype: ndarray
        """
        cur = self.__conn.cursor()
        cur.execute("SELECT X(GEOMETRY), Y(GEOMETRY), %s FROM noeuds ORDER BY OGC_FID"%(zColumn))
        return numpy.require(cur.fetchall(), numpy.float32, 'F')

    def nodeValues(self):
        """return values at nodes

        :return: values at nodes
        :rtype: ndarray
        """
        if not len(self.__nodeValues):
            return []
        else:
            return self.__unitConvertionFactor*self.__nodeValues[self.date()]

    def elementValues(self):
        """return values at elements

        :return: values at nodes
        :rtype: ndarray
        """
        if not len(self.__elementValues):
            return []
        else:
            return self.__unitConvertionFactor*self.__elementValues[self.date()]

    def maxValue(self, index=None):
        """ Return max mesh value 

        :param index: node index
        :type index: dict

        :return: max value
        :rtype: float
        """
        array = self.__nodeValues if not self.valueAtElement() else self.__elementValues
        if not len(array):
            return float('NaN')
        if index:
            return self.__unitConvertionFactor * numpy.max(array[index])
        else:
            return self.__unitConvertionFactor * max([max_ for max_ in \
                    [numpy.max(values) for values in array]])

    def minValue(self, index=None):
        """ Return min mesh value 

        :param index: node index
        :type index: dict

        :return: min value
        :rtype: float
        """
        array = self.__nodeValues if not self.valueAtElement() else self.__elementValues
        if not len(array):
            return float('NaN')
        if index:
            return self.__unitConvertionFactor * numpy.min(array[index])
        else:
            return self.__unitConvertionFactor * min([min_ for min_ in \
                    [numpy.min(values) for values in array]])

    def valuesAt(self, point, svalue=False):
        """return a list of values at a given point,
        interpolate values if needed, returns sigma values if required and available

        :param point: point xy
        :type point: tuple
        :param svalue: sigma values flag
        :type svalue: bool

        :return: value
        :rtype: float
        """
        cur = self.__conn.cursor()
        if not self.valueAtElement():
            element = []
            values = []
            svalues = []
            hassval = len(self.__nodeSValues)
            for nid in _getElementContaining(self.__conn, point):
                cur.execute("SELECT X(GEOMETRY), Y(GEOMETRY) FROM noeuds \
                        WHERE OGC_FID = "+str(nid))
                [x, y] = cur.fetchone()
                element.append((x, y))
                values.append(self.__nodeValues[:,nid-1])
                if hassval:
                    svalues.append(self.__nodeSValues[:,nid-1])
            values = self.__unitConvertionFactor*numpy.array(values)
            svalues = self.__unitConvertionFactor*numpy.array(svalues)

            weights = interpolationWeights(point, element)
            for i, w in enumerate(weights):
                values[i,:] *= w
                if hassval:
                    svalues[i,:] *= w
            if svalue:
                return (numpy.sum(values, axis=0), (numpy.sum(svalues, axis=0) if len(svalues) else []))
            else:
                return numpy.sum(values, axis=0)
        else:
            project_SRID = str(cur.execute("SELECT srid FROM geometry_columns WHERE f_table_name = 'noeuds'").fetchone()[0])
            eid = cur.execute("SELECT OGC_FID FROM mailles \
                    WHERE Intersects(GEOMETRY, MakePoint(%f, %f, %s)) \
                    LIMIT 1"%(point[0], point[1], project_SRID)).fetchone()
            if not eid:
                raise RuntimeError("Le point "+str(point)+" n'est pas dans le domaine")
            else:
                if svalue:
                    return (self.__elementValues[:, eid[0]-1]*self.__unitConvertionFactor,
                           (self.__elementSValues[:, eid[0]-1]*self.__unitConvertionFactor\
                            if len(self.__elementSValues) else []))
                else:
                    return self.__elementValues[:, eid[0]-1]*self.__unitConvertionFactor

    def measuresAt(self, point, withinDistance=.5):
        """returns a map of measures in the vicinity of the considered point
        the map as the mesure type as key and
        [(date, value, uncertainty)] as values

        :param point: point xy
        :type point: tuple
        :param withinDistance: snap distance
        :type withinDistance: float

        :return: measures
        :rtype: dict
        """
        if self.resultColumn() not in ['potentiel', 'concentration']:
            return {}
        measures = {}
        cur = self.__conn.cursor()
        project_SRID = str(cur.execute("SELECT srid FROM geometry_columns WHERE f_table_name = 'noeuds'").fetchone()[0])
        cur.execute("SELECT OGC_FID, nom FROM points_interet \
                WHERE PtDistWithin(GEOMETRY, MakePoint(%f, %f, %s), %f)"%(
                    point[0], point[1], project_SRID, withinDistance))
        for pid, nom in cur.fetchall():
            if self.resultColumn() == 'potentiel':
                cur.execute("SELECT type, date, potentiel, incertitude_potentiel \
                        FROM mesures WHERE pid=%d AND potentiel IS NOT NULL"%(pid))
            if self.resultColumn() == 'concentration':
                cur.execute("SELECT type, date, concentration, incertitude_concentration \
                        FROM mesures WHERE pid=%d AND concentration IS NOT NULL"%(pid))
            for type_, date, valeur, incertitude in cur.fetchall():
                if type_ not in measures:
                    measures[type_] = []
                format = "%Y-%m-%d %H:%M:%S" if ":" in date else "%Y-%m-%d"
                measures[type_].append((datetime.strptime(date, format),
                    valeur*self.__unitConvertionFactor, (incertitude*self.__unitConvertionFactor) if incertitude else 0))

        return measures

    def altitudeAt(self, point):
        """return altitude given point,
        interpolate values if needed

        :param point: point xy
        :type point: tuple

        :return: value
        :rtype: float
        """

        cur = self.__conn.cursor()
        element = []
        altitudes = []
        for nid in _getElementContaining(self.__conn, point):
            cur.execute("SELECT X(GEOMETRY), Y(GEOMETRY), altitude FROM noeuds \
                    WHERE OGC_FID = "+str(nid))
            [x, y, z] = cur.fetchone()
            element.append((x, y))
            altitudes.append(z)
        altitudes = numpy.array(altitudes)
        weights = interpolationWeights(point, element)
        for i, w in enumerate(weights):
            altitudes[i] *= w
        return numpy.sum(altitudes, axis=0)

    def mass_balance(self):
        """compute mass balance 

        :return sat: sat values
        :rtype sat: list
        :return out: out values
        :rtype out: list
        :return insat: insat values
        :rtype insat: list
        """
        cur = self.__conn.cursor()
        cur.execute("SELECT sat, out FROM bilan_masse ORDER BY did")
        values = numpy.require(cur.fetchall(), numpy.float32)
        if not len(values):
            return None, None, None
        sat = values[:,0]
        out = values[:,1]
        cur.execute("SELECT zns FROM bilan_masse WHERE zns IS NOT NULL ORDER BY did")
        insat = numpy.array(cur.fetchall(), numpy.float32)
        return sat, out, insat


if __name__ == "__main__":
    # test the spatialite data provider
    # USAGE: spatialitemeshdataprovider.py uri output image
    import sys



